"""
Design to be based on https://github.com/PyCQA/pylint/blob/master/pylint/lint.py

Return text based output
Also allow inlined comments

See http://python-jsonschema.readthedocs.io/en/latest/errors/ for error properties
"""


def report_error():
    pass
